"""Support testing with Pytest."""

import pytest
import os
import logging
from asgi_tools.tests import manage_lifespan

from . import TestClient


def pytest_addoption(parser):
    """Append pytest options for testing Muffin apps."""
    parser.addini('muffin_app', 'Set path to muffin application')
    parser.addoption('--muffin-app', dest='muffin_app', help='Set to muffin application')

    parser.addini('muffin_config', 'Set module path to muffin configuration')
    parser.addoption('--muffin-config', dest='muffin_config',
                     help='Set module path to muffin configuration')


def pytest_load_initial_conftests(early_config, parser, args):
    """Prepare to loading Muffin application."""
    from muffin import CONFIG_ENV_VARIABLE

    options = parser.parse_known_args(args)

    # Initialize configuration
    config = options.muffin_config or early_config.getini('muffin_config')
    if config:
        os.environ[CONFIG_ENV_VARIABLE] = config

    # Initialize application
    app_ = options.muffin_app or early_config.getini('muffin_app')
    early_config.app = app_


@pytest.fixture(scope='session')
async def app(pytestconfig, request, aiolib):
    """Load an application, run lifespan events, prepare plugins."""
    if not pytestconfig.app:
        logging.warning(
            'Improperly configured. Please set ``muffin_app`` in your pytest config. '
            'Or use ``--muffin-app`` command option.')
        return

    from muffin.utils import import_app

    app = import_app(pytestconfig.app)
    msg = f"Setup application '{app.cfg.name}'"
    if app.cfg.config:
        msg += f"with config '{app.cfg.config}'"
    app.logger.info(msg)

    async with manage_lifespan(app):

        # Setup plugins
        for plugin in app.plugins.values():
            if hasattr(plugin, 'conftest') and plugin.conftest is not None:
                app.logger.info(f"Setup plugin '{plugin.name}'")
                await plugin.conftest()

        yield app


@pytest.fixture
def client(app):
    """Generate a test client for the app."""
    return TestClient(app)



"""The Muffin Utils."""

import asyncio
import importlib
import pkgutil
import sys
import os
import typing as t
from types import ModuleType
import threading
from collections import OrderedDict

from asgi_tools._compat import trio, curio
from asgi_tools.typing import ASGIApp


__all__ = (
    'aio_lib', 'aio_run', 'import_submodules', 'current_async_library',
    'is_awaitable', 'to_awaitable'
)

AIOLIB = threading.local()
AIOLIB.current = None
AIOLIBS: t.Dict[str, ModuleType] = OrderedDict()

if curio:
    AIOLIBS['curio'] = curio

if trio:
    AIOLIBS['trio'] = trio

AIOLIBS['asyncio'] = asyncio


def aio_lib() -> str:
    """Return first available async library."""
    aiolib = os.environ.get('MUFFIN_AIOLIB', 'asyncio')
    if aiolib:
        return aiolib

    for name, module in AIOLIBS.items():
        if module is not None:
            return name

    return 'asyncio'


def aio_run(corofn: t.Callable[..., t.Awaitable], *args, **kwargs) -> t.Any:
    """Run the given coroutine with current async library."""
    AIOLIB.current = aiolib = AIOLIB.current or aio_lib()
    if aiolib == 'asyncio':
        return asyncio.run(corofn(*args, **kwargs))

    return AIOLIBS[aiolib].run(lambda: corofn(*args, **kwargs))  # type: ignore


def import_submodules(package_name: str, *submodules: str) -> t.Dict[str, ModuleType]:
    """Import all submodules by package name."""
    package = sys.modules[package_name]
    return {
        name: importlib.import_module(package_name + '.' + name)
        for _, name, _ in pkgutil.walk_packages(package.__path__)  # type: ignore # mypy #1422
        if not submodules or name in submodules
    }


def import_app(app_uri: str, reload: bool = False) -> t.Optional[ASGIApp]:
    """Import application by the given string (python.path.to.module:app_name)."""
    mod_name, _, app_name = app_uri.partition(':')
    mod = importlib.import_module(mod_name)
    if reload:
        importlib.reload(mod)

    try:
        return getattr(mod, app_name or 'app', None) or getattr(mod, 'application')
    except AttributeError:
        raise ImportError(f"Application {app_uri} not found.")


from sniffio import current_async_library               # noqa
from asgi_tools.utils import is_awaitable, to_awaitable # noqa



"""Muffin is a web framework to build ASGI applications."""

# Package information
# ===================

__version__ = "0.84.3"
__project__ = "muffin"
__author__ = "Kirill Klenov <horneds@gmail.com>"
__license__ = "MIT"

from asgi_tools import (  # noqa
    Request, Response, ResponseText, ResponseHTML, ResponseJSON, ResponseError,
    ResponseFile, ResponseRedirect, ResponseStream, ResponseSSE, ResponseWebSocket,
    ASGIMethodNotAllowed, ASGINotFound, ASGIConnectionClosed, ASGIError)
from asgi_tools.tests import ASGITestClient as TestClient  # noqa


class MuffinException(ASGIError):

    """Base class for Muffin Errors."""

    pass


CONFIG_ENV_VARIABLE = 'MUFFIN_CONFIG'


from .app import Application    # noqa
from .handler import Handler    # noqa
